status_dictionary = {
    0: 'Available',
    1: 'Selected',
    2: 'Active',
    3: 'Suspended',
    4: 'Completed',
    5: 'Problem',
    6: 'Not/Avl',
    7: 'Reset',
    8: 'Start',
    9: 'PWR Fail',
    10: 'No Cntllr',
    11: 'ShutDown',
    12: 'Starting',
    13: 'Blocked',
    14: 'Waiting',
    15: 'FRA',
    16: 'Rem Maint'
}
"""
Dictionary to decode the "result" field returned by the channel status messages.
"""